package core

import (
	"fmt"

	"git.urbach.dev/cli/q/src/errors"
	"git.urbach.dev/cli/q/src/expression"
	"git.urbach.dev/cli/q/src/ssa"
	"git.urbach.dev/cli/q/src/token"
	"git.urbach.dev/cli/q/src/types"
)

// Evaluate converts an expression to an SSA value.
func (f *Function) Evaluate(expr *expression.Expression) (ssa.Value, error) {
	if expr.IsLeaf() {
		switch expr.Token.Kind {
		case token.Identifier:
			name := expr.Token.String(f.File.Bytes)
			value, exists := f.Identifiers[name]

			if !exists {
				function, exists := f.All.Functions[f.File.Package+"."+name]

				if !exists {
					return nil, errors.New(&UnknownIdentifier{Name: name}, f.File, expr.Token.Position)
				}

				f.Dependencies.Add(function)

				v := &ssa.Function{
					UniqueName: function.UniqueName,
					Typ:        function.Type,
					IsExtern:   function.IsExtern(),
					Source:     ssa.Source(expr.Source),
				}

				return v, nil
			}

			return value, nil

		case token.Number:
			number, err := f.ToNumber(expr.Token)

			if err != nil {
				return nil, err
			}

			v := f.Append(&ssa.Int{
				Int:    number,
				Source: ssa.Source(expr.Source),
			})

			return v, nil

		case token.String:
			data := expr.Token.Bytes(f.File.Bytes)
			data = Unescape(data)

			length := f.Append(&ssa.Int{
				Int:    len(data),
				Source: ssa.Source(expr.Source),
			})

			pointer := f.Append(&ssa.Bytes{
				Bytes:  data,
				Source: ssa.Source(expr.Source),
			})

			v := &ssa.Struct{
				Arguments: []ssa.Value{pointer, length},
				Typ:       types.String,
				Source:    ssa.Source(expr.Source),
			}

			return v, nil
		}

		return nil, errors.New(InvalidExpression, f.File, expr.Token.Position)
	}

	switch expr.Token.Kind {
	case token.Call:
		if expr.Children[0].Token.Kind == token.Identifier && expr.Children[0].String(f.File.Bytes) == "syscall" {
			args, err := f.Decompose(expr.Children[1:], nil)

			if err != nil {
				return nil, err
			}

			syscall := &ssa.Syscall{
				Arguments: args,
				Source:    ssa.Source(expr.Source),
			}

			return f.Append(syscall), nil
		}

		funcValue, err := f.Evaluate(expr.Children[0])

		if err != nil {
			return nil, err
		}

		ssaFunc := funcValue.(*ssa.Function)
		fn := f.All.Functions[ssaFunc.UniqueName]
		inputExpressions := expr.Children[1:]

		if len(inputExpressions) != len(fn.Input) {
			return nil, errors.New(&ParameterCountMismatch{Function: fn.UniqueName, Count: len(inputExpressions), ExpectedCount: len(fn.Input)}, f.File, expr.Source[0].Position)
		}

		args, err := f.Decompose(inputExpressions, fn.Input)

		if err != nil {
			return nil, err
		}

		if fn.IsExtern() {
			v := f.Append(&ssa.CallExtern{Call: ssa.Call{
				Func:      ssaFunc,
				Arguments: args,
				Source:    ssa.Source(expr.Source),
			}})

			return v, nil
		}

		v := f.Append(&ssa.Call{
			Func:      ssaFunc,
			Arguments: args,
			Source:    ssa.Source(expr.Source),
		})

		return v, nil

	case token.Dot:
		left := expr.Children[0]
		right := expr.Children[1]
		leftText := left.String(f.File.Bytes)
		rightText := right.String(f.File.Bytes)
		fullName := fmt.Sprintf("%s.%s", leftText, rightText)
		identifier, exists := f.Identifiers[fullName]

		if exists {
			return identifier, nil
		}

		function, exists := f.All.Functions[fullName]

		if exists {
			if function.IsExtern() {
				f.Assembler.Libraries = f.Assembler.Libraries.Append(function.Package, function.Name)
			} else {
				f.Dependencies.Add(function)
			}

			v := &ssa.Function{
				UniqueName: function.UniqueName,
				Typ:        function.Type,
				IsExtern:   function.IsExtern(),
				Source:     ssa.Source(expr.Source),
			}

			return v, nil
		}

		return nil, errors.New(&UnknownIdentifier{Name: fullName}, f.File, left.Token.Position)

	default:
		if expr.Token.IsOperator() {
			left := expr.Children[0]
			right := expr.Children[1]

			leftValue, err := f.Evaluate(left)

			if err != nil {
				return nil, err
			}

			rightValue, err := f.Evaluate(right)

			if err != nil {
				return nil, err
			}

			v := f.Append(&ssa.BinaryOp{
				Left:   leftValue,
				Right:  rightValue,
				Op:     expr.Token.Kind,
				Source: ssa.Source(expr.Source),
			})

			return v, nil
		}

		panic("not implemented")
	}
}